function output_vec = gen_diff(input_vec , diff_ord , OA_flag)

% output_vec = gen_diff(input_vec , diff_ord , OA_flag)
%
% The function implements the generalized finite difference of order 
% "diff_ord" and its adjoint. The boundary condition is assumed to be all
% zero, e.g., 
%       gen_diff([a b c] , 1 , 1) = [a b-a  c-b   ]     and
%       gen_diff([a b c] , 2 , 1) = [a b-2a c-2b+a]
%
% "input_vec" is the vector which is subject to the finite difference or
% its adjoint operator.
%
% "diff_ord" is the order of finite difference operator.
%
% "OA_flag" is a flag accepting the values 1 and -1 which represents the
% mode of the function:
%       OA_flag = 1     => Original operator
%       OA_flag = -1    => Adjoint  operator 


% converting the input to a row vector
if size(input_vec , 1) > 1
    x   = input_vec.';
else
    x   = input_vec;
end



% implementation of the operator
if OA_flag == 1
    
    % first order operator
    D   = @(x) [x(1)    x(2:end)-x(1:end-1)];
    
    output_vec  = D(x);
    for aa = 1 : diff_ord-1
        output_vec  = D(output_vec);
    end
    
else
    
    % first order operator
    DT  = @(x) [x(1:end-1)-x(2:end)  x(end)];
    
    output_vec  = DT(x);
    for aa = 1 : diff_ord-1
        output_vec  = DT(output_vec);
    end
end


% reshaping the vector to the original form
output_vec      = reshape(output_vec , size(input_vec));