% The objective of this subroutine is to evaluate the average performance
% of various methods for a range of input noise variances. 
% Note that for regularization techniques, the coefficient 'lambda' is the
% one given by the user. The program initially performs 10 test runs and
% optimizes the parameter lambda. Then, the average of those 10 obtained 
% lambda's will be used for the main minimizations. 



handles.Cmpltd_Perc = 0;    % completed percentage of the performance comparison




% parameters and initializations
N_noiseVar              = 10;
handles.noiseVar_vec    = logspace( log10(handles.AWGN_varL) , log10(handles.AWGN_varU) , N_noiseVar);
N_testRun               = 10;
D                       = @(x,OA) gen_diff(x , handles.Operator , OA);
l                       = 4^handles.Operator;       % Lipshitz constant




% Initialization for the generator function
auxillary       = handles;
if handles.Dist_Sel == 3    % Poisson innovation
    auxillary.Dist_Sel      = handles.P_Dist_Sel;
    auxillary.Prior_Type    = 2;
    
elseif handles.Dist_Sel > 3
    auxillary.Dist_Sel      = auxillary.Dist_Sel - 1;
    auxillary.Prior_Type    = 1;
end




% Initialization for the MMSE denoiser (if enabled)
if handles.MethodEn(1) == 1     % MMSE
    prior.param     = handles.Dist_Param;
    switch handles.Dist_Sel
        case 1  % Gaussian
            prior.type  = 'g';
        case 2  % Laplace
            prior.type  = 'l';
        case 3  % Poisson
            prior.rho   = 1 - handles.MassProb;
            switch handles.P_Dist_Sel
                case 1  % Poisson-Gaussian
                    prior.type  = 'pg';
                case 2  % Poisson-Laplace
                    prior.type  = 'pl';
                case 3  % Poisson-StudentT
                    prior.type  = 'ps';
            end
        case 4  % Student's-t
            prior.type  = 's';
        case 5  % alpha-stable
            prior.type  = 'a';
    end
end
    




% Defining the outputs
if handles.MethodEn(1) == 1     % MMSE
    handles.Av_ImpSNR_MMSE  = zeros(1 , N_noiseVar);
end
if handles.MethodEn(2) == 1     % LMMSE
    handles.Av_ImpSNR_LMMSE = zeros(1 , N_noiseVar);
    handles.Ave_LMMSE_lambda= zeros(1 , N_noiseVar);
end
if handles.MethodEn(3) == 1     % Log
    handles.Av_ImpSNR_Log   = zeros(1 , N_noiseVar);
    handles.Ave_Log_lambda  = zeros(1 , N_noiseVar);
end
if handles.MethodEn(4) == 1     % L1
    handles.Av_ImpSNR_L1    = zeros(1 , N_noiseVar);
    handles.Ave_L1_lambda   = zeros(1 , N_noiseVar);
end



% run counters
N_TotRun    = (handles.NumRealz * sum(handles.MethodEn) + N_testRun * sum(handles.MethodEn(2:end)) ) * N_noiseVar;
N_Run       = 0;
set(handles.Compl_Percentage_Text   , 'String' , sprintf('%3.2f %%', N_Run/N_TotRun*100) )


%%%%%%%%%%
%%% Main iterations
%%%%%%%%%%%%%%%%%%%%%%%%%%%
noiseVar_ind    = 0;
%for noiseVar_ind = 1 : N_noiseVar
while ( noiseVar_ind < N_noiseVar)& isequal(get(handles.Compare_PushButton , 'String') , 'Stop') %( handles.Cmpr_state )
    noiseVar_ind    = noiseVar_ind + 1;
    auxillary.Noise_Var     = handles.noiseVar_vec(noiseVar_ind);
    
    % test runs to find optimum lambda value
    if sum(handles.MethodEn(2:4)) > 0
        test_Run    = 0;
        %for test_Run = 1: N_testRun
        while ( test_Run < N_testRun)& isequal(get(handles.Compare_PushButton , 'String') , 'Stop')
            test_Run    = test_Run + 1;
            [SS_data , SS_dataN]    = priorGen(auxillary);

            
            % optimum lambda for LMMSE
            if ( handles.MethodEn(2) )& isequal(get(handles.Compare_PushButton , 'String') , 'Stop')   % LMMSE
                MSE_LMMSE       = @(lambda) mean( (SS_data - CG_L2(SS_dataN , lambda , D)) .^ 2 );
                lambda_opt      = fminbnd(MSE_LMMSE , 0 , 10 * auxillary.Noise_Var);
                %lambda_opt      = Golden_Section_Search(MSE_LMMSE , 0 , 10 * auxillary.Noise_Var);
                
                handles.Ave_LMMSE_lambda(noiseVar_ind)  = ( (test_Run-1) * handles.Ave_LMMSE_lambda(noiseVar_ind) + lambda_opt ) / test_Run;
            
                % progress indicator
                N_Run           = N_Run + 1;
                set(handles.Compl_Percentage_Text   , 'String' , sprintf('%3.2f %%', N_Run/N_TotRun*100) )
                pie([N_Run/N_TotRun   1-N_Run/N_TotRun] , [1 0])
                colormap winter;
                getframe;
            end
            
            % optimum lambda for Log
            if ( handles.MethodEn(3) )& isequal(get(handles.Compare_PushButton , 'String') , 'Stop')   % Log
                MSE_Log         = @(lambda) mean( (SS_data - ADMM_log(SS_dataN , lambda , D , handles.delta)) .^ 2 );
                lambda_opt      = fminbnd(MSE_Log , 0 , 10 * auxillary.Noise_Var);
                %lambda_opt      = Golden_Section_Search(MSE_Log , 0 , 10 * auxillary.Noise_Var);
                
                handles.Ave_Log_lambda(noiseVar_ind)    = ( (test_Run-1) * handles.Ave_Log_lambda(noiseVar_ind) + lambda_opt ) / test_Run;
                
                % progress indicator
                N_Run           = N_Run + 1;
                set(handles.Compl_Percentage_Text   , 'String' , sprintf('%3.2f %%', N_Run/N_TotRun*100) )
                pie([N_Run/N_TotRun   1-N_Run/N_TotRun] , [1 0])
                colormap winter;
                getframe;
            end
            
            % optimum lambda for L1
            if ( handles.MethodEn(4) )& isequal(get(handles.Compare_PushButton , 'String') , 'Stop')   % L1
                MSE_L1          = @(lambda) mean( (SS_data - Duality_L1(SS_dataN , lambda , D , 'Lipschitz' , l)) .^ 2 );
                lambda_opt      = fminbnd(MSE_L1 , 0 , 10 * auxillary.Noise_Var);
                %lambda_opt      = Golden_Section_Search(MSE_L1 , 0 , 10 * auxillary.Noise_Var);
                
                handles.Ave_L1_lambda(noiseVar_ind)     = ( (test_Run-1) * handles.Ave_L1_lambda(noiseVar_ind) + lambda_opt ) / test_Run;
                
                % progress indicator
                N_Run           = N_Run + 1;
                set(handles.Compl_Percentage_Text   , 'String' , sprintf('%3.2f %%', N_Run/N_TotRun*100) )
                pie([N_Run/N_TotRun   1-N_Run/N_TotRun] , [1 0])
                colormap winter;
                getframe;
            end
            
            
        end
    end
    
    
    % main runs
    main_Run        = 0;
    %for main_Run = 1: handles.NumRealz
    while ( main_Run < handles.NumRealz)& isequal(get(handles.Compare_PushButton , 'String') , 'Stop')
        main_Run    = main_Run + 1;
        [SS_data , SS_dataN]    = priorGen(auxillary);
        SNR                     = 10 * log10(mean(SS_data.^2) / mean( (SS_dataN - SS_data).^2 ));
        
        % MSE improvement for MMSE
        if ( handles.MethodEn(1) )& isequal(get(handles.Compare_PushButton , 'String') , 'Stop')   % MMSE
            % Sampling grid
            xmax    = ceil(max(abs(SS_dataN)) + 2 * sqrt(handles.Dist_Param + auxillary.Noise_Var));
            delta   = min(handles.Dist_Param , auxillary.Noise_Var);
            imax    = ceil(xmax/delta - 0.5);
            sgrid   = (-imax : imax) * delta;
    
            % denoising and MSE calculation 
            SNR_MMSE        = 10 * log10( mean(SS_data.^2) / mean( (SS_data - TMP_MMSE_Levy(SS_dataN , sgrid , prior , auxillary.Noise_Var)) .^ 2 ) );
            handles.Av_ImpSNR_MMSE(noiseVar_ind)    = ( (main_Run-1) * handles.Av_ImpSNR_MMSE(noiseVar_ind) + (SNR_MMSE-SNR) ) / main_Run;
            
            % progress indicator
            N_Run           = N_Run + 1;
            set(handles.Compl_Percentage_Text   , 'String' , sprintf('%3.2f %%', N_Run/N_TotRun*100) )
            pie([N_Run/N_TotRun   1-N_Run/N_TotRun+1e-9] , [1 0])
            colormap winter;
            getframe;
        end
        
        % MSE improvement for LMMSE
        if ( handles.MethodEn(2) )& isequal(get(handles.Compare_PushButton , 'String') , 'Stop')   % LMMSE
            lambda          = handles.Ave_LMMSE_lambda(noiseVar_ind);
            SNR_LMMSE       = 10 * log10( mean(SS_data.^2) / mean( (SS_data - CG_L2(SS_dataN , lambda , D)) .^ 2 ));
            handles.Av_ImpSNR_LMMSE(noiseVar_ind)  = ( (main_Run-1) * handles.Av_ImpSNR_LMMSE(noiseVar_ind) + (SNR_LMMSE-SNR) ) / main_Run;
            
            % progress indicator
            N_Run           = N_Run + 1;
            set(handles.Compl_Percentage_Text   , 'String' , sprintf('%3.2f %%', N_Run/N_TotRun*100) )
            pie([N_Run/N_TotRun   1-N_Run/N_TotRun+1e-9] , [1 0])
            colormap winter;
            getframe;
        end
        
        % MSE improvement for Log
        if ( handles.MethodEn(3) )& isequal(get(handles.Compare_PushButton , 'String') , 'Stop')   % Log
            lambda          = handles.Ave_Log_lambda(noiseVar_ind);
            SNR_Log         = 10 * log10( mean(SS_data.^2) / mean( (SS_data - ADMM_log(SS_dataN , lambda , D , handles.delta)) .^ 2 ));
            handles.Av_ImpSNR_Log(noiseVar_ind)     = ( (main_Run-1) * handles.Av_ImpSNR_Log(noiseVar_ind) + (SNR_Log-SNR) ) / main_Run;
            
            % progress indicator
            N_Run           = N_Run + 1;
            set(handles.Compl_Percentage_Text   , 'String' , sprintf('%3.2f %%', N_Run/N_TotRun*100) )
            pie([N_Run/N_TotRun   1-N_Run/N_TotRun+1e-9] , [1 0])
            colormap winter;
            getframe;
        end
        
        % MSE improvement for L1
        if ( handles.MethodEn(4) )& isequal(get(handles.Compare_PushButton , 'String') , 'Stop')   % L1
            lambda          = handles.Ave_L1_lambda(noiseVar_ind);
            SNR_L1          = 10 * log10( mean(SS_data.^2) / mean( (SS_data - Duality_L1(SS_dataN , lambda , D , 'Lipschitz' , l)) .^ 2 ));
            handles.Av_ImpSNR_L1(noiseVar_ind)      = ( (main_Run-1) * handles.Av_ImpSNR_L1(noiseVar_ind) + (SNR_L1-SNR) ) / main_Run;
            
            % progress indicator
            N_Run           = N_Run + 1;
            set(handles.Compl_Percentage_Text   , 'String' , sprintf('%3.2f %%', N_Run/N_TotRun*100) )
            pie([N_Run/N_TotRun   1-N_Run/N_TotRun+1e-4] , [1 0])
            colormap winter;
            getframe;
        end
        
    end

    
end






%%%%%%%%%%%%%%
%%% Plotting the results
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
if isequal(get(handles.Compare_PushButton , 'String') , 'Stop')
    results     = zeros(sum(handles.MethodEn) , N_noiseVar);
    legend_text = repmat(' ' , sum(handles.MethodEn) , 5);
    res_ind     = 0;
    
    if handles.MethodEn(1) == 1     % MMSE
        res_ind     = res_ind + 1;
        results(res_ind , :)    = handles.Av_ImpSNR_MMSE;
        legend_text(res_ind , :)= 'MMSE ';
    end
    if handles.MethodEn(2) == 1     % LMMSE
        res_ind     = res_ind + 1;
        results(res_ind , :)    = handles.Av_ImpSNR_LMMSE;
        legend_text(res_ind , :)= 'LMMSE';
    end
    if handles.MethodEn(3) == 1     % Log
        res_ind     = res_ind + 1;
        results(res_ind , :)    = handles.Av_ImpSNR_Log;
        legend_text(res_ind , :)= ' Log ';
    end
    if handles.MethodEn(4) == 1     % L1
        res_ind     = res_ind + 1;
        results(res_ind , :)    = handles.Av_ImpSNR_L1;
        legend_text(res_ind , :)= ' L1  ';
    end
    
    semilogx(handles.noiseVar_vec , results)
    ylabel('\Delta SNR [dB]')
    xlabel('\sigma_n^2')
    legend(legend_text)
    grid on
end